# coding: utf-8

"""
    Info:
    ~~~~~
    to debug the output in a browser: add "?debug=1" to a url.
    To reformat the payload: add "?debug=2" to a url.
    (This works only if settings.DEBUG==True)
"""

import sys
import struct
import logging

try:
    from functools import wraps
except ImportError:
    from django.utils.functional import wraps  # Python 2.3, 2.4 fallback.

#try:
#    import json # New in Python v2.6
#except ImportError:
#    from django.utils import simplejson as json

try:
    import json
except ImportError:
    try:
        import simplejson as json
    except ImportError:
        sys.stderr.write("Error: json (or simplejson) module is needed")
        sys.exit(1)

from django.conf import settings
from django.http import HttpResponse

from weave_project.weave.utils import timestamp, datetime2epochtime, cut, WeaveTimestamp

logging.basicConfig(level=logging.DEBUG)


def assert_username(debug=False):
    """ Check if the username from the url is logged in. """
    def _inner(view_function):
        @wraps(view_function)
        def _check_user(request, *args, **kwargs):
            user = request.user

            if not user.is_authenticated():
                msg = "Permission denied for anonymous user. Please log in."
                if debug:
                    logging.debug(msg)
                raise PermissionDenied(msg)

            username = kwargs["username"]
            if user.username != username:
                msg = "Wrong user!"
                if debug:
                    msg += " (%r != %r)" % (user.username, username)
                    logging.debug(msg)
                raise PermissionDenied("Wrong user!")

            return view_function(request, *args, **kwargs)
        return _check_user
    return _inner


# monkey patch for round all floats in json response
#def float_repr(f):
#    # XXX: Is this really needed???
#    return "%.2f" % round(f, 2)
#json.encoder.FLOAT_REPR = float_repr


def json_dumps(data, **extra):
    data_string = json.dumps(data, sort_keys=True, separators=(',', ':'), **extra)
#    data_string = data_string.replace("/", "\\/") # XXX: Is this really needed???
    return data_string


def json_response(debug=False):
    """
    A view can return tuple, list or dict:
        tupe: (item, weave timestamp) - item would be serialized
        list/dict: the weave timestamp would be added.
        
    TODO: handle "X-If-Unmodified-Since" header
    """
    def renderer(function):
        @wraps(function)
        def wrapper(request, *args, **kwargs):
            response = function(request, *args, **kwargs)

            if isinstance(response, HttpResponse):
                return response

            if isinstance(response, tuple):
                data, weave_timestamp = response
            elif isinstance(response, (dict, list)):
                data = response
                weave_timestamp = timestamp()
            else:
                msg = (
                    "json_response info: %s has not return tuple, dict or list, has return: %r (%r)"
                ) % (function.__name__, type(data), function.func_code)
                raise AssertionError(msg)

            if settings.DEBUG and "debug" in request.GET:
                logging.debug("debug output for %r:" % function.__name__)
                content_type = "text/plain"

                if int(request.GET["debug"]) > 1:
                    def load_payload(item):
                        if "payload" in item:
                            raw_payload = item["payload"]
                            payload_dict = json.loads(raw_payload)
                            item["payload"] = payload_dict
                        return item

                    if isinstance(data, list):
                        data = [load_payload(item) for item in data]
                    else:
                        data = load_payload(data)

                response_string = json_dumps(data, indent=4)

            else:
                accept = request.META.get('HTTP_ACCEPT', 'application/json')
#                if accept == "application/whoisi":
#                    content_type = "application/whoisi"
#                    response_string = struct.pack('!I', data.id) + json.dumps(data)
                if accept == "application/newlines":
                    content_type = "application/newlines"
                    if isinstance(data, list):
                        output = []
                        for obj in data:
                            output.append(json_dumps(obj))
                        response_string = "\n".join(output) # FIXME: '\n' -> r'\u000a' ???
                    else:
                        response_string = json_dumps(data)
                else:
                    content_type = "application/json"
                    response_string = json_dumps(data)

            response = HttpResponse(response_string, content_type=content_type)

            # https://wiki.mozilla.org/Labs/Weave/Sync/1.0/API#X-Weave-Timestamp
#            response["X-Weave-Timestamp"] = "%.2f" % weave_timestamp
            response["X-Weave-Timestamp"] = weave_timestamp

            # https://wiki.mozilla.org/Labs/Weave/Sync/1.0/API#X-Weave-Alert
#            response["X-Weave-Alert"] = "render debug for %r:" % function.__name__

            return response
        return wrapper
    return renderer



# coding:utf-8

"""
    :copyleft: 2010 by the django-weave team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

import logging
try:
    import json # New in Python v2.6
except ImportError:
    from django.utils import simplejson as json

from django.db import models
from django.conf import settings
from django.contrib.auth.models import User
from django.contrib.sites.models import Site
from django.utils.translation import ugettext as _
from django.contrib.sites.managers import CurrentSiteManager

# http://code.google.com/p/django-tools/
from django_tools.middlewares import ThreadLocal

# django-weave own stuff
from utils import timestamp, datetime2epochtime


logging.basicConfig(level=logging.DEBUG)


class BaseModel(models.Model):
    createtime = models.DateTimeField(auto_now_add=True, help_text="Create time")
    lastupdatetime = models.DateTimeField(auto_now=True, help_text="Time of the last change.")

    createby = models.ForeignKey(User, editable=False, related_name="%(class)s_createby",
        null=True, blank=True, # <- If the model used outsite a real request (e.g. unittest, db shell)
        help_text="User how create this entry.")
    lastupdateby = models.ForeignKey(User, editable=False, related_name="%(class)s_lastupdateby",
        null=True, blank=True, # <- If the model used outsite a real request (e.g. unittest, db shell)
        help_text="User as last edit this entry.")

    def save(self, *args, **kwargs):
        """
        Automatic update createby and lastupdateby attributes with the request object witch must be
        the first argument.
        """
        current_user = ThreadLocal.get_current_user()

        if current_user and isinstance(current_user, User):
            if self.pk == None or kwargs.get("force_insert", False): # New model entry
                self.createby = current_user
            self.lastupdateby = current_user

        return super(BaseModel, self).save(*args, **kwargs)

    class Meta:
        abstract = True


class Lock(BaseModel):
    user = models.ForeignKey(User)
    lock_path = models.CharField(max_length=255)
#    lock_type = models.CharField(max_length=255)
#    lock_scope = models.CharField(max_length=255)
#    depth = models.CharField(max_length=255)
#    timeout = models.PositiveIntegerField()

    class Meta:
        unique_together = ("user", "lock_path")



class CollectionManager(CurrentSiteManager):
    def get_or_create2(self, user, col_name):
        collection, created = self.get_or_create(
            user=user, name=col_name
        )
        if created:
            logging.info("Collection %r created" % collection)
            current_site = Site.objects.get_current()
            collection.sites.add(current_site)
            collection.save()
        else:
            logging.info("Collection %r exists" % collection)

        return collection


class Collection(BaseModel):
    """
    https://wiki.mozilla.org/Labs/Weave/Sync/1.0/Setup
    
    inherited from BaseModel:
        createtime     -> datetime of creation
        lastupdatetime -> datetime of the last change
        createby       -> ForeignKey to user who creaded this entry
        lastupdateby   -> ForeignKey to user who has edited this entry
    """
    user = models.ForeignKey(User)
    name = models.CharField(max_length=96)

    sites = models.ManyToManyField(Site, default=[settings.SITE_ID])
    on_site = CollectionManager('sites')

    def site_info(self):
        """ for admin.ModelAdmin list_display """
        sites = self.sites.all()
        return ", ".join([site.name for site in sites])
    site_info.short_description = _('Exists on site')
    site_info.allow_tags = False

    def __unicode__(self):
        return u"%r (user %r)" % (self.name, self.user.username)

    class Meta:
        ordering = ("-lastupdatetime",)


class WboManager(models.Manager):
    def create_or_update(self, payload_dict, collection, user):
        """
        create or update a wbo
        TODO:
            - Check parentid, but how?
            - must wboid + parentid be unique?
        """
        wbo, created = Wbo.objects.get_or_create(
            collection=collection, user=user, wboid=payload_dict["id"],
            defaults={
                "parentid": payload_dict.get("parentid", None),
                "predecessorid": payload_dict.get("predecessorid", None),
                "sortindex": payload_dict.get("sortindex", None),
                "modified": payload_dict.get("modified", timestamp()),
                "payload": payload_dict["payload"],
            }
        )
        if created:
            msg = "New wbo %r created" % wbo.wboid
        else:
            wbo.parentid = payload_dict.get("parentid", None)
            wbo.predecessorid = payload_dict.get("predecessorid", None)
            wbo.sortindex = payload_dict.get("sortindex", None)
            wbo.modified = payload_dict.get("modified", timestamp())
            wbo.payload = payload_dict["payload"]
            wbo.save()
            msg = "Existing wbo %r updated" % wbo.wboid

        logging.info("%s (collection: %r, user: %r)" % (msg, collection.name, user.username))
        return wbo, created



class Wbo(BaseModel):
    """
    https://wiki.mozilla.org/Labs/Weave/Sync/1.0/API   
    https://wiki.mozilla.org/Labs/Weave/Sync/1.0/Setup
    
    inherited from BaseModel:
        createtime     -> datetime of creation
        lastupdatetime -> datetime of the last change
        createby       -> ForeignKey to user who creaded this entry
        lastupdateby   -> ForeignKey to user who has edited this entry
    """
    objects = WboManager()

    collection = models.ForeignKey(Collection, blank=True, null=True)
    user = models.ForeignKey(User)
    wboid = models.CharField(max_length=64, blank=True,
        help_text="wbo identifying string"
    )
    parentid = models.CharField(max_length=64, blank=True, null=True,
        help_text="wbo parent identifying string"
    )
    predecessorid = models.CharField(max_length=64, blank=True, null=True,
        help_text="wbo predecessorid"
    )
    sortindex = models.IntegerField(null=True, blank=True,
        help_text="An integer indicting the relative importance of this item in the collection.",
    )
    modified = models.FloatField(help_text="X-Weave-Timestamp")
    payload = models.TextField(blank=True,
        help_text=(
            "A string containing a JSON structure encapsulating the data of the record."
            " This structure is defined separately for each WBO type. Parts of the"
            " structure may be encrypted, in which case the structure should also"
            " specify a record for decryption."
        )
    )
    def get_response_dict(self):
        response_dict = {
            "id": self.wboid,
            "modified": self.modified,
            "payload": self.payload,
        }
        for key in ("parentid", "predecessorid", "sortindex"):
            value = getattr(self, key)
            if value:
                response_dict[key] = value
        return response_dict

    def __unicode__(self):
        return u"%r (%r)" % (self.wboid, self.collection)

    class Meta:
        ordering = ("-lastupdatetime",)



# coding: utf-8

"""
    weave_project.settings
    ~~~~~~~~~~~~~~~~~~~~~~

    IMPORTANT:
        You should not edit this file!
        Owerwrite settings with a local settings file:
            local_settings.py

    Here are not all settings predifined you can use. Please look at the
    django documentation for a full list of all items:
        http://www.djangoproject.com/documentation/settings/

    Last commit info:
    ~~~~~~~~~~~~~~~~~
    $LastChangedDate:$
    $Rev:$
    $Author: JensDiemer $

    :copyleft: 2010 by the django-weave team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

import os
import sys

try:
    #from django_tools.utils import info_print;info_print.redirect_stdout()
    import django
    import weave
except Exception, e:
    import traceback
    print "-" * 79
    sys.stderr.write(traceback.format_exc())
    print "-" * 79
    raise


WEAVE_PROJECT_ROOT = os.path.abspath(os.path.dirname(__file__))

DEBUG = True
TEMPLATE_DEBUG = DEBUG

ADMINS = (
    # ('Your Name', 'your_email@domain.com'),
)

MANAGERS = ADMINS

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.', # Add 'postgresql_psycopg2', 'postgresql', 'mysql', 'sqlite3' or 'oracle'.
        'NAME': '', # Or path to database file if using sqlite3.
        'USER': '', # Not used with sqlite3.
        'PASSWORD': '', # Not used with sqlite3.
        'HOST': '', # Set to empty string for localhost. Not used with sqlite3.
        'PORT': '', # Set to empty string for default. Not used with sqlite3.
    }
}

# Django can't handling time zone very good.
# The Django default TIME_ZONE is 'America/Chicago' (Central Standard Time Zone, (CST), UTC-6)
# but this is not the best choice.
# We set it to "UTC" (same as Greenwich Mean Time, GMT-0, without daylight-saving time)
# All datetime (e.g. model createtime) would be stored in UTC.
# See also: http://groups.google.com/group/django-developers/browse_thread/thread/4ca560ef33c88bf3
TIME_ZONE = "UTC"

# Language code for this installation. All choices can be found here:
# http://www.i18nguy.com/unicode/language-identifiers.html
LANGUAGE_CODE = 'en-us'

SITE_ID = 1

# If you set this to False, Django will make some optimizations so as not
# to load the internationalization machinery.
USE_I18N = True

# Absolute path to the directory that holds media.
# Example: "/home/media/media.lawrence.com/"
MEDIA_ROOT = ''

# URL that handles the media served from MEDIA_ROOT. Make sure to use a
# trailing slash if there is a path component (optional in other cases).
# Examples: "http://media.lawrence.com", "http://example.com/media/"
MEDIA_URL = ''

# URL prefix for admin media -- CSS, JavaScript and images. Make sure to use a
# trailing slash.
# Examples: "http://foo.com/media/", "/media/".
ADMIN_MEDIA_PREFIX = '/media/'

# Make this unique, and don't share it with anybody.
SECRET_KEY = "Make this unique, and don't share it with anybody!"

# List of callables that know how to import templates from various sources.
TEMPLATE_LOADERS = (
    'django.template.loaders.filesystem.Loader',
    'django.template.loaders.app_directories.Loader',
#     'django.template.loaders.eggs.Loader',
)

MIDDLEWARE_CLASSES = (
    'weave.middlewares.debug.DebugMiddleware',

    'django.middleware.common.CommonMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.transaction.TransactionMiddleware',

    'django_tools.middlewares.ThreadLocal.ThreadLocalMiddleware',
    'reversion.middleware.RevisionMiddleware',
)

ROOT_URLCONF = 'weave_project.urls'


TEMPLATE_DIRS = (
    # Put strings here, like "/home/html/django_templates" or "C:/www/django/templates".
    # Always use forward slashes, even on Windows.
    # Don't forget to use absolute paths, not relative paths.

    os.path.join(os.path.abspath(os.path.dirname(weave.__file__)), "templates"),
    os.path.normpath(os.path.join(WEAVE_PROJECT_ROOT, "../../reversion/reversion/templates")),
    os.path.join(os.path.abspath(os.path.dirname(django.__file__)), "contrib/admin/templates"),
)

for dir in TEMPLATE_DIRS:
    if not os.path.isdir(dir):
        print "TEMPLATE_DIR %r doesn't exist!" % dir

INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.sites',
    'django.contrib.messages',
    'django.contrib.admin',
    'django.contrib.admindocs',

    'reversion',
    "weave",
)

try:
    from local_settings import *
except ImportError:
    pass



# coding: utf-8

import time
import datetime

MAX = 120

def timestamp():
    # Weave rounds to 2 digits and so must we, otherwise rounding errors will
    # influence the "newer" and "older" modifiers
    return round(time.time(), 2)


def datetime2epochtime(dt):
    assert isinstance(dt, datetime.datetime)
    timestamp = time.mktime(dt.timetuple()) # datetime -> time since the epoch
    # Add microseconds. FIXME: Is there a easier way?
    timestamp += (dt.microsecond / 1000000.0)
#    print "round: %r" % timestamp,
    timestamp = round(timestamp, 2)
#    print "to %r" % timestamp
    return timestamp


def cut(s, max=None):
    if max is None:
        max = MAX
    s = repr(s)
    if len(s) > max:
        return s[:max] + "..."
    return s

class WeaveTimestamp(object):
    def __init__(self, timestamp):
        assert isinstance(timestamp, float)
        self.timestamp = timestamp

    def to_json(self):
        print self.timestamp
        r = round(self.timestamp, 2)
        print r
        return r

    def __repr__(self):
        return u"<WeaveTimestamp %r>" % self.timestamp




# coding: utf-8

import os
import sys
import inspect
import traceback

from django.http import Http404
from django.conf import settings
from django.utils import termcolors
from django.contrib.redirects.models import Redirect
from django.core.management.color import color_style
from django.core import urlresolvers

from weave_project.weave.utils import cut


STACK_LIMIT = 60

SKIP_URLS = (
    urlresolvers.reverse("admin:index"),
)
# Activate only debug, if this string is in the url, debug every request, if None:
#DEBUG_ONLY = "pubkey"
DEBUG_ONLY = None


#def print_stack_info():
#    """
#    return stack_info: Where from the announcement comes?
#    """
#    self_basename = os.path.basename(__file__)
#    if self_basename.endswith(".pyc"):
#        # cut: ".pyc" -> ".py"
#        self_basename = self_basename[:-1]
#
#    stack_list = inspect.stack()
#    for stack_line in reversed(stack_list):
#        filename = cut(stack_line[1])
#        lineno = stack_line[2]
#        func_name = stack_line[3]
#        print "%s %4s %s" % (filename, lineno, func_name)
#
##    # go forward in the stack, to outside of this file.
##    for no, stack_line in enumerate(stack_list):
##        filename = stack_line[1]
##        if os.path.basename(filename) != self_basename:
##            break
##
##    stack_list = stack_list[no:no + STACK_LIMIT] # limit the displayed stack info
##
##    stack_info = []
##    for stack_line in reversed(stack_list):
##        filename = cut(stack_line[1])
##        lineno = stack_line[2]
##        func_name = stack_line[3]
##        print "%s %4s %s" % (filename, lineno, func_name)


def skip_debug(request):
    if DEBUG_ONLY is not None and DEBUG_ONLY not in request.path:
        return True # No debugging

    path = request.path
    for skip_url in SKIP_URLS:
        if path.startswith(skip_url):
            return True # No debugging
    return False


class DebugMiddleware(object):
    style = color_style()

    def process_view(self, request, view_func, view_args, view_kwargs):
        if skip_debug(request):
            return

        print "*** DebugMiddleware.process_view():"
#        for i in dir(view_func):print i, getattr(view_func, i, "---")
        print "view: %s,  args: %r, kwagrs: %r" % (
            self.style.NOTICE(view_func.__name__), view_args, view_kwargs
        )

    def process_request(self, request):
        if skip_debug(request):
            return

        print "-" * 79
        print "*** DebugMiddleware.process_request():"
        print self.style.SQL_TABLE(request.method), self.style.SQL_FIELD(request.build_absolute_uri())
        print "request.META['CONTENT_LENGTH']: %r" % request.META['CONTENT_LENGTH']
        for k, v in sorted(request.META.iteritems()):
            upper_key = k.upper()
            if upper_key.startswith("HTTP_ACCEPT"):
                print k, self.style.HTTP_NOT_MODIFIED(v)
            elif upper_key.startswith("HTTP") or upper_key.startswith("X-"):
                print k, v
        print "request.GET: %r" % request.GET
        print "request.POST: %s" % cut(request.POST)
        print "request.FILES: %r" % request.FILES
        print "request.raw_post_data: %s" % cut(request.raw_post_data)
#        print_stack_info()
#        print sys.exc_info()

    def process_exception(self, request, exception):
        print self.style.NOTICE("*** DebugMiddleware.process_exception():")
#        print_stack_info()
        message = cut(str(exception))
        print message
        print traceback.format_exc()

    def process_response(self, request, response):
        if skip_debug(request):
            return response

        print "*** DebugMiddleware.process_response():"
#        print_stack_info()
#        print sys.exc_info()
        print "response: %r" % response
        print "response headers: %r" % response._headers
        status_code = response.status_code
        status_code_msg = "response.status_code: %r" % response.status_code
        if status_code == 200:
            print self.style.SQL_KEYWORD(status_code_msg)
        else:
            print self.style.NOTICE(status_code_msg)

        print "response.content: %s" % cut(response.content)
#        print "response.content:"
#        print response.content
#        print "-" * 79
        with file("last_error_page.html", "w") as f:
            f.write(response.content)
        return response



# coding: utf-8

"""
    :copyleft: 2010 by the django-weave team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

import time
import httplib
import logging
import datetime
import traceback
from xml.etree import cElementTree as ET
from decimal import Decimal

try:
    import json # New in Python v2.6
except ImportError:
    from django.utils import simplejson as json


from django.conf import settings
from django.contrib.auth.models import User
from django.core.urlresolvers import reverse
from django.contrib.sites.models import Site
from django.shortcuts import render_to_response
from django.template.loader import render_to_string
from django.core.exceptions import PermissionDenied
from django.views.decorators.csrf import csrf_exempt
from django.http import HttpResponseBadRequest, HttpResponse

# http://code.google.com/p/django-tools/
from django_tools.decorators import check_permissions, render_to

# django-weave own stuff
from weave_project.weave.models import Lock, Collection, Wbo
from weave_project.weave.forms import ChangePasswordForm
from weave_project.weave.utils import timestamp, datetime2epochtime, WeaveTimestamp
from weave_project.weave.decorators import json_response


# from http://hg.mozilla.org/labs/weave/file/tip/tools/scripts/weave_server.py#l189
ERR_UID_OR_EMAIL_AVAILABLE = "1"
ERR_WRONG_HTTP_METHOD = "-1"
ERR_MISSING_UID = "-2"
ERR_INVALID_UID = "-3"
ERR_UID_OR_EMAIL_IN_USE = "0"
ERR_EMAIL_IN_USE = "-5"
ERR_MISSING_PASSWORD = "-8"
ERR_MISSING_RECAPTCHA_CHALLENGE_FIELD = "-6"
ERR_MISSING_RECAPTCHA_RESPONSE_FIELD = "-7"
ERR_MISSING_NEW = "-11"
ERR_INCORRECT_PASSWORD = "-12"
ERR_ACCOUNT_CREATED_VERIFICATION_SENT = "2"
ERR_ACCOUNT_CREATED = "3"


logging.basicConfig(level=logging.DEBUG)
#logging.debug('This is a debug message')
#logging.info('This is an info message')
#logging.warning('This is a warning message')
#logging.error('This is an error message')
#logging.critical('This is a critical error message')




class WeaveResponse(HttpResponse):
    def __init__(self, content='', status=None, content_type=None, weave_timestamp=None):
        super(WeaveResponse, self).__init__(content=content, status=status, content_type=content_type)

        if weave_timestamp is None:
            weave_timestamp = timestamp()
#        self["X-Weave-Timestamp"] = "%.2f" % weave_timestamp
        self["X-Weave-Timestamp"] = weave_timestamp


class PlaintextResponse(WeaveResponse):
    """ plaintext response with a status code in the content """
    def __init__(self, content="", status=None, weave_timestamp=None):
        super(PlaintextResponse, self).__init__(
            content=content, status=status, content_type="text/plain", weave_timestamp=weave_timestamp
        )


def render_xml_response(template_name, context, status=httplib.OK, weave_timestamp=None):
    rendered = render_to_string(template_name, context)
    response = WeaveResponse(
        content=rendered, content_type="text/xml", status=status, weave_timestamp=weave_timestamp
    )
    return response



class RecordNotFoundResponse(WeaveResponse):
    status_code = 404
    def __init__(self, msg, content):
        HttpResponse.__init__(self)
        self["X-Weave-Alert"] = msg
        self._container = [content]


#@render_to('weave/root_view.html')
@json_response(debug=True)
def root_view(request):
    print " *** root_view! ***"
    return {}


def wbo_timestamp_reponse(queryset):
    wbos = queryset.values("wboid", "modified")

    newest = 0
    timestamps = {}
    for wbo in wbos:
        modified = wbo["modified"]
        if modified > newest:
            newest = modified
        timestamps[wbo["wboid"]] = modified
#        timestamps[wbo["wboid"]] = "%.2f" % modified
#        timestamps[wbo["wboid"]] = WeaveTimestamp(modified)

    return timestamps, newest


#@assert_username(debug=True)
#@check_permissions(superuser_only=False, permissions=(u'weave.add_collection', u'weave.add_wbo'))
@json_response(debug=True)
def info_collections(request, version, username):
    """
    return a list of all exsiting collections with the newest wbo timestamp
    https://wiki.mozilla.org/Labs/Weave/Sync/1.0/API#GET
    
    e.g.: {"clients":1268646032.25,"crypto":1268149025.85,"forms":1268149016.51,"history":...}
    """
    user = User.objects.get(username=username)

    collections = Collection.on_site.filter(user=user)

    response_dict = {}
    for collection in collections:
        queryset = Wbo.objects.filter(collection=collection)
        queryset = queryset.order_by("-modified")
        queryset = queryset.values_list("modified", flat=True)
        newest_timestamp = queryset[0]
        response_dict[collection.name] = newest_timestamp

    return response_dict



def limit_wbo_queryset(request, queryset):
    """
    TODO:
    predecessorid = fromform(form, "predecessorid")
    full = fromform(form, "full")
    """
    GET = request.GET

    ids = GET.get("ids", None)
    if ids is not None:
#        ids = [id.strip() for id in ids.split(",") if id.strip()] 
        ids = ids.split(",")
        logging.debug("limit wbo queryset with ids: %r" % ids)
        queryset = queryset.filter(wboid__in=ids)

    parentid = GET.get("parentid", None)
    if parentid is not None:
        queryset = queryset.filter(parentid=parentid)

    newer = GET.get("newer", None)
    if newer is not None: # Greater than or equal to newer modified timestamp
        newer = float(newer)
        queryset = queryset.filter(modified__gte=newer)

    older = GET.get("older", None)
    if older is not None: # Less than or equal to older modified timestamp
        older = float(older)
        queryset = queryset.filter(modified__lte=older)

    index_above = GET.get("index_above", None)
    if index_above is not None: # Greater than or equal to index_above modified timestamp
        index_above = int(index_above)
        queryset = queryset.filter(sortindex__gte=index_above)

    index_below = GET.get("index_below", None)
    if index_below is not None: # Less than or equal to index_below modified timestamp
        index_below = int(index_below)
        queryset = queryset.filter(sortindex__lte=index_below)

    sort_type = GET.get("sort", None)
    if sort_type is not None:
        if sort_type == 'oldest':
            queryset = queryset.order_by("modified")
        elif sort_type == 'newest':
            queryset = queryset.order_by("-modified")
        elif sort_type == 'index':
            queryset = queryset.order_by("wboid")
        else:
            raise NameError("sort type %r unknown" % sort_type)

    offset = GET.get("offset", None)
    if offset is not None:
        offset = int(offset)
        queryset = queryset[offset:]

    limit = GET.get("limit", None)
    if limit is not None:
        limit = int(limit)
        queryset = queryset[:limit]

    return queryset



#@assert_username(debug=True)
#@check_permissions(superuser_only=False, permissions=(u'weave.add_collection', u'weave.add_wbo'))
@csrf_exempt
@json_response(debug=True)
def storage(request, version, username, col_name, wboid=None):
    """ Get/add/remove items from weave storage. """

#    user = request.user
    user = User.objects.get(username=username)

    if request.method == 'GET':
        # Returns a list of the WBO contained in a collection.
        try:
            collection = Collection.on_site.get(user=user, name=col_name)
        except Collection.DoesNotExist:
            msg = "Collection %r for user %r not found" % (col_name, username)
            logging.info(msg)
            if "full" in request.GET:
                return []
            else:
                return RecordNotFoundResponse(msg, content="")
        logging.debug("get collection %r" % collection)

        wbo_queryset = Wbo.objects.all().filter(user=user).filter(collection=collection)

        if wboid is not None: # return one WBO
            logging.debug("Use wboid %r to filter the queryset" % wboid)
            wbo_queryset = wbo_queryset.filter(wboid=wboid)
            qs_count = wbo_queryset.count()
            if qs_count == 0:
                msg = "wbo %r not found in collection %r" % (wboid, col_name)
                logging.debug(msg)
                return RecordNotFoundResponse(msg, content="")
            elif qs_count == 1:
                wbo = wbo_queryset[0]
                logging.debug("return only one WBO: %r" % wbo)
                response_dict = wbo.get_response_dict()
                return response_dict, wbo.modified
            logging.debug("more than one wbo found: queryset count: %r" % qs_count)
        else:
            wbo_queryset = limit_wbo_queryset(request, wbo_queryset)

            # If defined, returns the full WBO, rather than just the id. 
            full = request.GET.get("full", None)
            if full is None:
                logging.debug("GET parameter 'full' not set -> return only the WBO ids")
                return wbo_timestamp_reponse(wbo_queryset)

        newest_timestamp = 0
        wbo_list = []
        for wbo in wbo_queryset:
            if wbo.modified > newest_timestamp:
                newest_timestamp = wbo.modified
            response_dict = wbo.get_response_dict()
            wbo_list.append(response_dict)
        logging.info("Return a wbo list with %d items." % len(wbo_list))
        return wbo_list, newest_timestamp

    elif request.method in ("POST", "PUT"):
        # https://wiki.mozilla.org/Labs/Weave/Sync/1.0/API#PUT
        # Adds the WBO defined in the request body to the collection.
        raw_payload = request.raw_post_data
        if not raw_payload:
            # If the WBO does not contain a payload, it will only update
            # the provided metadata fields on an already defined object.
            raise NotImplemented

        collection = Collection.on_site.get_or_create2(user, col_name)

        payload = json.loads(raw_payload)

        if isinstance(payload, list):
            if request.method == "PUT":
                logging.error("PUT request with a payload array???")

            newest_timestamp = 0
            success_list = []
            for item in payload:
                # TODO: Check "invalid parentid", but how?
                wbo, created = Wbo.objects.create_or_update(item, collection, user)
                success_list.append(wbo.wboid)
                if wbo.modified > newest_timestamp:
                    newest_timestamp = wbo.modified

            # Returns a hash of successful and unsuccessful saves, including guidance as to possible errors: 
            # {"modified":1233702554.25,
            #   "success":["{GXS58IDC}12","{GXS58IDC}13","{GXS58IDC}15","{GXS58IDC}16","{GXS58IDC}18","{GXS58IDC}19"],
            #   "failed":{"{GXS58IDC}11":["invalid parentid"],"{GXS58IDC}14":["invalid parentid"],"{GXS58IDC}17":["invalid parentid"],"{GXS58IDC}20":["invalid parentid"]}
            # }
            response_dict = {"modified": newest_timestamp, "success":success_list, "failed":{}}
            return response_dict, newest_timestamp
        else:
            if request.method == "POST":
                logging.error("POST request without a payload array???")

            assert payload["id"] == wboid, "wrong wbo id: %r != %r" % (payload["id"], wboid)
            wbo, created = Wbo.objects.create_or_update(payload, collection, user)

            # return the timestamp associated with the modification.
            response_dict = {wboid: wbo.modified}
            return response_dict, wbo.modified

    elif request.method == 'DELETE':
        # The DELETE method seems not to work in this way, it's documented:
        # https://wiki.mozilla.org/Labs/Weave/Sync/1.0/API#DELETE
        try:
            collection = Collection.on_site.get(user=user, name=col_name)
        except Collection.DoesNotExist:
            logging.info("%r is not a collection key, try to delete a wbo" % col_name)
            wboid = col_name
        else:
            logging.info("Delete %r and all wbos in this collection for user %r" % (col_name, username))
            wbo_queryset = Wbo.objects.filter(user=user).filter(collection=collection)
            wbo_queryset = limit_wbo_queryset(request, wbo_queryset)
            logging.info("Delete wbo: %r" % wbo_queryset)
            wbo_queryset.delete()
            logging.info("Delete Collection: %r" % collection)
            collection.delete()
            return WeaveResponse()

        try:
            wbo = Wbo.objects.filter(user=user).get(wboid=wboid)
        except Wbo.DoesNotExist:
            msg = "Wbo %r not exist in collection %r for user %r" % (wboid, col_name, username)
            logging.info(msg)
            return RecordNotFoundResponse(msg, content="")
        logging.info("Delete wbo %r in collection %r for user %r" % (wboid, col_name, username))
        wbo.delete()
        return WeaveResponse()

    else:
        raise NotImplementedError("request.method %r not implemented!" % request.method)




def handle_lock(request, username, lock_path=None):
    """
    Is this really used by Firefox weave plugin?!?!
    """
    try:
        user = User.objects.get(username=username)
    except User.DoesNotExist:
        print "User %r doesn't exist!" % username
        return PlaintextResponse(ERR_INVALID_UID)

    print "request.raw_post_data: %r" % request.raw_post_data

    if request.method == 'PROPFIND':
        if not lock_path:
            locks = Lock.objects.filter(user=user)
            items = [{"path": lock.lock_path, "prop": ""} for lock in locks]
            context = {"items": items}
            return render_xml_response("weave/propfind.xml", context, status=httplib.MULTI_STATUS)
        else:
            try:
                lock = Lock.objects.get(user=user, lock_path=lock_path)
            except Lock.DoesNotExist:
    #            count = Lock.objects.filter(user=user).count()
    #            token = "opaquelocktoken:%d" % (count + 1)
                context = {
                    "items": [{"path": lock_path, "prop": ""}]
                }
                return render_xml_response("weave/propfind.xml", context, status=httplib.MULTI_STATUS)

            context = {"token_id": lock.id}
            return render_xml_response("weave/lock.xml", context)
    elif request.method == 'LOCK':
        lock, created = Lock.objects.get_or_create(user=user, lock_path=lock_path)
        if created:
            lock.save()
            context = {"token_id": lock.id}
            return render_xml_response("weave/lock.xml", context)
        else:
            return HttpResponse(status=httplib.LOCKED)

#        raw_post_data = request.raw_post_data
#        elementtree = ET.XML(raw_post_data)
#        print elementtree
#        token = elementtree.find(".//{DAV:}locktoken/{DAV:}href").text
#        print token
#        print "XXX"
#        context = {"token_id": None}
#        return render_xml_response("weave/lock.xml", context)



    elif request.method == 'UNLOCK':
        token = request.META["HTTP_LOCK_TOKEN"]
        print "***", token

        try:
            lock = Lock.objects.get(user=user, lock_path=lock_path)
        except Lock.DoesNotExist:
            print "lock %r doesn't exist for user %r" % (lock_path, user)
            #return WeaveResponse(status=httplib.BAD_REQUEST)
        else:
            lock.delete()

        return WeaveResponse(status=httplib.NO_CONTENT)

    raise

#    def _handle_UNLOCK(self, path):
#        
#        if path not in self.locks:
#            return HttpResponse(httplib.BAD_REQUEST)
#        if token == "<%s>" % self.locks[path]:
#            del self.locks[path]
#            return HttpResponse(httplib.NO_CONTENT)
#        return HttpResponse(httplib.BAD_REQUEST)


    raise

#        return PlaintextResponse(status=httplib.LOCKED)


@csrf_exempt
def chpwd(request):
    """
    Change the user password.
    """
    if request.method != 'POST':
        logging.error("wrong request method %r" % request.method)
        return HttpResponseBadRequest()

    form = ChangePasswordForm(request.POST)
    if not form.is_valid():
        # TODO
        print "*** Form error:"
        print form.errors
        ERR_MISSING_UID
        ERR_MISSING_PASSWORD
        raise #FIXME return HttpResponseBadRequest(status=)

    username = form.cleaned_data["uid"]
    password = form.cleaned_data["password"] # the old password
    new = form.cleaned_data["new"] # the new password

    try:
        user = User.objects.get(username=username)
    except User.DoesNotExist:
        logging.debug("User %r doesn't exist!" % username)
        return PlaintextResponse(ERR_INVALID_UID)

    if user.check_password(password) != True:
        logging.debug("Old password %r is wrong!" % password)
        return PlaintextResponse(ERR_INCORRECT_PASSWORD, httplib.BAD_REQUEST)
    else:
        logging.debug("Old password %r is ok." % password)

    user.set_password(new)
    user.save()
    logging.debug("Password for User %r changed from %r to %r" % (username, password, new))
    return PlaintextResponse()



def sign_in(request, version, username):
    """
    finding cluster for user -> return 404 -> Using serverURL as data cluster (multi-cluster support disabled)
    """
    assert version == "1"

    try:
        User.objects.get(username=username)
    except User.DoesNotExist:
        logging.debug("User %r doesn't exist!" % username)
        return PlaintextResponse(ERR_UID_OR_EMAIL_AVAILABLE)
    else:
        logging.debug("User %r exist." % username)

        return PlaintextResponse(ERR_UID_OR_EMAIL_IN_USE, status="404")


@csrf_exempt
def register_check(request, username):
    """
    Returns 1 if the username exist, 0 if not exist.
    """
    try:
        User.objects.get(username=username)
    except User.DoesNotExist:
        logging.debug("User %r doesn't exist!" % username)
        return PlaintextResponse(ERR_UID_OR_EMAIL_AVAILABLE)
    else:
        logging.debug("User %r exist." % username)
        return PlaintextResponse(ERR_UID_OR_EMAIL_IN_USE)



@csrf_exempt
def exist_user(request, version, username):
    """
    https://wiki.mozilla.org/Labs/Weave/User/1.0/API
    Returns 1 if the username is in use, 0 if it is available.
    
    e.g.: https://auth.services.mozilla.com/user/1/UserName
    """
    try:
        User.objects.get(username=username)
    except User.DoesNotExist:
        logging.debug("User %r doesn't exist!" % username)
        return PlaintextResponse("0")
    else:
        logging.debug("User %r exist." % username)
        return PlaintextResponse("1")


def captcha_html(request, version):
    print "_" * 79
    print "captcha_html:"
#    raise Http404
    #response = HttpResponse("11", status=400, content_type='application/json')
    response = HttpResponse("not supported")
    response["X-Weave-Timestamp"] = timestamp()
    return response


@json_response(debug=True)
def setup_user(request, version, username):
    absolute_uri = request.build_absolute_uri()
    return {}, timestamp()





from django import forms

class ChangePasswordForm(forms.Form):
    uid = forms.CharField(max_length='30')
    password = forms.CharField(max_length='128')
    new = forms.CharField(max_length='128')



# coding: utf-8

"""
    PyLucid.admin
    ~~~~~~~~~~~~~~

    Register all PyLucid model in django admin interface.

    Last commit info:
    ~~~~~~~~~~~~~~~~~
    $LastChangedDate$
    $Rev$
    $Author$

    :copyleft: 2010 by the django-weave team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


from django.contrib import admin
from django.conf import settings
from django.contrib.auth.models import User, Permission
from django.contrib.auth.admin import UserAdmin

from reversion.admin import VersionAdmin

from weave.models import Lock, Wbo, Collection

class LockAdmin(VersionAdmin):
    list_display = ("id", "lastupdatetime", "user", "lock_path")
    date_hierarchy = 'lastupdatetime'

admin.site.register(Lock, LockAdmin)


class WboAdmin(VersionAdmin):
    def payload_cutout(self, obj):
        MAX = 100
        payload = obj.payload
        if len(payload) > MAX:
            payload = payload[:MAX] + "..."
        return payload
    payload_cutout.short_description = "Payload cutout"
#    view_on_site_link.allow_tags = True

    list_display = ("id", "lastupdatetime", "modified", "user", "wboid", "parentid", "predecessorid", "sortindex", "lastupdateby", "payload_cutout")
    list_display_links = ("wboid",)
    list_filter = ("user", "collection", "predecessorid", "parentid")
    date_hierarchy = 'lastupdatetime'
    search_fields = ("wboid", "parentid", "predecessorid", "sortindex", "payload")

admin.site.register(Wbo, WboAdmin)


class CollectionAdmin(VersionAdmin):
    """
    """
    list_display = ("id", "lastupdatetime", "user", "name", "lastupdateby")
    list_display_links = ("name",)
    list_filter = ("user", "createby", "lastupdateby",)
    date_hierarchy = 'lastupdatetime'
#    search_fields = ("headline", "content")

admin.site.register(Collection, CollectionAdmin)


if settings.DEBUG:
    from reversion.models import Revision, Version

    class RevisionAdmin(admin.ModelAdmin):
        list_display = ("id", "date_created", "user", "comment")
        list_display_links = ("date_created",)
        date_hierarchy = 'date_created'
        ordering = ('-date_created',)
        list_filter = ("user", "comment")
        search_fields = ("user", "comment")

    admin.site.register(Revision, RevisionAdmin)


    class VersionAdmin(admin.ModelAdmin):
        list_display = ("object_repr", "revision", "object_id", "content_type", "format",)
        list_display_links = ("object_repr", "object_id")
        list_filter = ("content_type", "format")
        search_fields = ("object_repr", "serialized_data")

    admin.site.register(Version, VersionAdmin)





# coding: utf-8

"""
    :copyleft: 2010 by the django-weave team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


from django.conf.urls.defaults import patterns, url

from weave import views


urlpatterns = patterns('',
    url(r'^(?P<version>.*?)/(?P<username>.*?)/storage/(?P<col_name>.*?)/(?P<wboid>.*?)$', views.storage),
    url(r'^(?P<version>.*?)/(?P<username>.*?)/storage/(?P<col_name>.*?)$', views.storage),

    url(r'^(?P<version>.*?)/(?P<username>.*?)/info/collections$',
        views.info_collections, name='info_collections'),

    url(r'^misc/(?P<version>.*?)/captcha_html$', views.captcha_html, name='captcha_html'),

    url(r'^user/(?P<version>.*?)/(?P<username>.*?)/node/weave$', views.sign_in, name='sign_in'),
    url(r'^user/(?P<version>.*?)/(?P<username>.*?)$', views.exist_user, name='exist_user'),

    url(r'^user/(?P<username>.*?)/(?P<lock_path>.*?)$', views.handle_lock, name='handle_lock'),

    url(r'^api/register/chpwd/$', views.chpwd, name='chpwd'),
    url(r'^api/register/check/(?P<username>.*?)$', views.register_check, name='register_check'),

    url(r'^', views.root_view, name='root_view'),
)
